import random


class DeriveVocab:


    def __init__(self, **kwargs):
        self.vocab = None

        self.original_data = None

        self.master_rankings = None

        self.derived_rankings = None

        self.similarities = None

        self.input_ranking = 3




    def initial_rankings(self, data=None):
        data = self.original_data = data if data else self.test_data()
        rankings = {}
        similarities = {}

        for item in data["input"]:
            rankings[list(item.keys())[0]] = list(item.values())[0]["rating"]
            try:
                similarities[list(item.keys())[0]] = list(item.values())[0]["data"]["attributes"]
            except:
                similarities[list(item.keys())[0]] = list(item.values())[0]["data"][0]["attributes"]

        self.master_rankings = rankings
        self.similarities = similarities
        self.get_vocab(rankings)
        self.derive_rankings()

        return self.vocab
        

    def get_vocab(self, rankings, length=None, data=None):
        self.vocab = self.sort_rankings(rankings)[:self.original_data["length"] if not length else length]
        return self.vocab
        

    def current_vocab(self):
        try:
            vocab = [list(i.values())[0]["data"][0]["item"] for i in self.original_data["input"] if [j for j in self.vocab if j == list(i.keys())[0]]]
        except:
            vocab = [list(i.values())[0]["data"]["item"] for i in self.original_data["input"] if [j for j in self.vocab if j == list(i.keys())[0]]]
        return vocab
    

    def sort_ratings(self, rankings):
        data = {}
        sort = sorted(rankings, key=lambda x: rankings[x], reverse=True)
        for item in sort:
            data[item] = rankings[item]
        return data
    

    def sort_rankings(self, rankings):
        rankings = sorted(rankings, key=lambda x: rankings[x], reverse=True)
        return rankings
    

    def derive_rankings(self):
        rankings = self.master_rankings

        for item in self.vocab:
            rankings[item] += len(rankings) - self.vocab.index(item)
            [self.add_rankings(rankings, i, abs(self.vocab.index(item)-len(self.vocab))) for i in self.similarities[item]] 

        self.master_rankings = rankings
        return rankings
    

    def post_derived(self):
        for item in self.original_data["input"]:
            item[list(item.keys())[0]]["rating"] += round(sum([self.master_rankings[i] for i in item[list(item.keys())[0]]["data"]["attributes"]])/len(item[list(item.keys())[0]]["data"]["attributes"]))
        self.sort_data(self.original_data)
        for item in self.original_data["input"]:
            item[list(item.keys())[0]].pop("rating", None)

        return self.original_data
    

    def get_master(self):
        return self.original_data
    

    def sort_data(self, data):
        sort = sorted(data["input"], key=lambda x: list(x.values())[0]["rating"], reverse=True)
        return sort
    

    def add_rankings(self, rankings, i, add):
        try:
            rankings[i] = rankings[i] + add
        except:
            pass


    def create_dataset(self, input, data):
        rankings = {}
        for item in data["input"]:
            rankings[list(item.keys())[0]] = 0

        for item in input:
            adjust = [list(i.keys())[0] for i in data["input"] if self.shared_attributes(item, i[list(i.keys())[0]]["data"]["attributes"]) == True]
            rankings = self.fill_rankings(rankings, adjust, item, input[item])

        rankings = self.sort_ratings(rankings)
        return rankings


    def fill_rankings(self, rankings, adjust, item, value):
        rankings[item] += 1
        for fill in adjust:
            try:
                rankings[fill] += 1
            except:
                pass
        for rank in rankings:
            try:
                if rank in adjust:
                    rankings[rank] = rankings[rank] + 1 if value == True else rankings[rank] - 1
            except:
                pass
        rankings.pop(item, None) if value == None else None
        return rankings


    def shared_attributes(self, input, data):
        item = list([i for i in self.original_data["input"] if list(i.keys())[0] == input][0].items())[0][1]["data"]["attributes"]
        return len([i for i in item if i in data]) > 0 

   
    def adjust_rankings(self, input, data=None):
        self.original_data = data if data else self.original_data
        rankings = self.create_dataset(input, data)
        self.master_rankings = rankings
        return self.get_vocab(rankings)
    

    # the original dataset (rating_initial.json in example data folder). it is what will actually be outputted by the core engine. this can be any type of data, but needs to be in this format. the actual values being used in the output are in the 'data'->'item' key of each item. the attributes are intended to be used for all similar items and can be used to adjust the vocab after the initial list has been built. for the vocab output, items with the top ratings will be chosen and the size is based on the value of the length key in the dataset.
    def test_data(self):
        return {
            "input": [
                {"strength_training": {"rating": 85, "data": {"item": "kettlebell_lift.png", "attributes": ["power_exercise","conditioning_workout"]}}},
                {"endurance_run": {"rating": 74, "data": {"item": "trail_run.png", "attributes": ["stamina_build","marathon_training"]}}},
                {"sprint_speed": {"rating": 72, "data": {"item": "sprint_track.png", "attributes": ["timing_practice","agility_course"]}}},
                {"agility_course": {"rating": 70, "data": {"item": "obstacle_course.png", "attributes": ["coordination_drill","mobility_drill"]}}},
                {"balance_challenge": {"rating": 66, "data": {"item": "tightrope.png", "attributes": ["control_drill","flexibility_session"]}}},
                {"coordination_drill": {"rating": 69, "data": {"item": "drumming.png", "attributes": ["rhythm_training","timing_practice"]}}},
                {"reaction_test": {"rating": 72, "data": {"item": "ping_pong.png", "attributes": ["coordination_drill","focus_practice"]}}},
                {"precision_training": {"rating": 76, "data": {"item": "sniper_training.png", "attributes": ["focus_practice","strategy_game"]}}},
                {"strategy_game": {"rating": 69, "data": {"item": "war_sim.png", "attributes": ["precision_training","teamwork_drill"]}}},
                {"teamwork_drill": {"rating": 72, "data": {"item": "relay_race.png", "attributes": ["stamina_build","leadership_task"]}}},
                {"leadership_task": {"rating": 71, "data": {"item": "team_coaching.png", "attributes": ["teamwork_drill","strategy_game"]}}},
                {"focus_practice": {"rating": 78, "data": {"item": "meditation.png", "attributes": ["precision_training","reaction_test"]}}},
                {"power_exercise": {"rating": 86, "data": {"item": "clap_pushup.png", "attributes": ["strength_training","conditioning_workout"]}}},
                {"flexibility_session": {"rating": 63, "data": {"item": "contortion.png", "attributes": ["balance_challenge","mobility_drill"]}}},
                {"stamina_build": {"rating": 73, "data": {"item": "rowing.png", "attributes": ["endurance_run","conditioning_workout"]}}},
                {"control_drill": {"rating": 71, "data": {"item": "hoverboard.png", "attributes": ["balance_challenge","flexibility_session"]}}},
                {"timing_practice": {"rating": 70, "data": {"item": "drum_circle.png", "attributes": ["sprint_speed","coordination_drill"]}}},
                {"rhythm_training": {"rating": 68, "data": {"item": "dance_routine.png", "attributes": ["coordination_drill","focus_practice"]}}},
                {"conditioning_workout": {"rating": 74, "data": {"item": "crossfit_box.png", "attributes": ["power_exercise","stamina_build"]}}},
                {"mobility_drill": {"rating": 62, "data": {"item": "foam_rolling.png", "attributes": ["agility_course","flexibility_session"]}}},
                {"deadlift_session": {"rating": 88, "data": {"item": "deadlift.png", "attributes": ["strength_training","power_exercise"]}}},
                {"marathon_training": {"rating": 76, "data": {"item": "marathon.png", "attributes": ["endurance_run","stamina_build"]}}},
                {"hurdle_run": {"rating": 71, "data": {"item": "hurdles.png", "attributes": ["agility_course","sprint_speed"]}}},
                {"soccer_drill": {"rating": 71, "data": {"item": "soccer_dribble.png", "attributes": ["teamwork_drill","stamina_build"]}}},
                {"surfing_practice": {"rating": 66, "data": {"item": "surf_board.png", "attributes": ["balance_challenge","mobility_drill"]}}},
                {"volleyball_training": {"rating": 69, "data": {"item": "volleyball_spike.png", "attributes": ["teamwork_drill","coordination_drill"]}}},
                {"boxing_reaction": {"rating": 71, "data": {"item": "boxing_reaction.png", "attributes": ["reaction_test","stamina_build"]}}},
                {"dart_practice": {"rating": 79, "data": {"item": "dart_throw.png", "attributes": ["precision_training","focus_practice"]}}},
                {"board_strategy": {"rating": 77, "data": {"item": "board_game.png", "attributes": ["strategy_game","precision_training"]}}},
                {"soccer_teamwork": {"rating": 73, "data": {"item": "soccer_game.png", "attributes": ["teamwork_drill","leadership_task"]}}},
                {"orchestra_lead": {"rating": 72, "data": {"item": "orchestra_conductor.png", "attributes": ["leadership_task","focus_practice"]}}},
                {"archery_focus": {"rating": 80, "data": {"item": "archery_target.png", "attributes": ["precision_training","focus_practice"]}}},
                {"vertical_jump": {"rating": 84, "data": {"item": "vertical_jump.png", "attributes": ["power_exercise","strength_training"]}}},
                {"pilates_session": {"rating": 61, "data": {"item": "pilates.png", "attributes": ["flexibility_session","mobility_drill"]}}},
                {"tennis_endurance": {"rating": 72, "data": {"item": "tennis_match.png", "attributes": ["stamina_build","coordination_drill"]}}},
                {"skateboard_control": {"rating": 71, "data": {"item": "skateboard_trick.png", "attributes": ["control_drill","agility_course"]}}},
                {"boxing_timing": {"rating": 69, "data": {"item": "boxing_combo.png", "attributes": ["timing_practice","reaction_test"]}}},
                {"jump_rope_rhythm": {"rating": 68, "data": {"item": "jump_rope_routine.png", "attributes": ["rhythm_training","coordination_drill"]}}},
                {"stair_conditioning": {"rating": 73, "data": {"item": "stair_climb.png", "attributes": ["conditioning_workout","stamina_build"]}}},
                {"hip_mobility": {"rating": 60, "data": {"item": "hip_mobility.png", "attributes": ["mobility_drill","flexibility_session"]}}},
                {"bench_press": {"rating": 86, "data": {"item": "bench_press.png", "attributes": ["strength_training","power_exercise"]}}},
                {"swimming_marathon": {"rating": 75, "data": {"item": "swimming_lap.png", "attributes": ["endurance_run","stamina_build"]}}},
                {"bike_sprint": {"rating": 72, "data": {"item": "bike_sprint.png", "attributes": ["sprint_speed","timing_practice"]}}},
                {"basketball_drill": {"rating": 70, "data": {"item": "basketball_dribble.png", "attributes": ["coordination_drill","teamwork_drill"]}}},
                {"snowboard_balance": {"rating": 65, "data": {"item": "snowboard.png", "attributes": ["balance_challenge","control_drill"]}}},
                {"cycling_marathon": {"rating": 74, "data": {"item": "cycling_marathon.png", "attributes": ["endurance_run","stamina_build"]}}},
                {"acrobatics_control": {"rating": 74, "data": {"item": "acrobatics.png", "attributes": ["agility_course","control_drill"]}}},
                {"soccer_tackle": {"rating": 70, "data": {"item": "soccer_tackle.png", "attributes": ["soccer_drill","coordination_drill"]}}},
                {"rowing_rhythm": {"rating": 68, "data": {"item": "rowing_stroke.png", "attributes": ["rhythm_training","stamina_build"]}}},
                {"boxing_training": {"rating": 75, "data": {"item": "boxing_training.png", "attributes": ["stamina_build","conditioning_workout"]}}},
                {"quarterback_drill": {"rating": 72, "data": {"item": "quarterback_drill.png", "attributes": ["leadership_task","teamwork_drill"]}}}
            ],
            "rules": "rating",
            "length": 5
        }

    

    # this is an example of the format needed to adjust the selected vocab. it assumes these are the current vocab items being used by the core engine. a true value indicates that the item and items with related attributes have a lower preference for the next output cycle. a false rating will emphasize that the item and its related items have a higher preference ranking. these values can be determined in whatever way is deemed necessary within the app the app they are being used in(such as user input, elapsed time, etc). 
    def input_format(self):
        return {
            "deadlift_session": False,
            "power_exercise": False,
            "bench_press": False,
            "strength_training": False,
            "vertical_jump": False,
        }


    def test_input(self):
        input = {}
        for item in self.vocab:
            input[item] = random.choice([True, False, True, False, True, False, None])
        return input
    

    def test(self, tests=10):
        print("\n\n\nINITIAL DATA\n")
        self.initial_rankings(self.test_data())
        self.original_data = self.sort_ratings(self.master_rankings)
        print(self.original_data)
        print("\n\nINITIAL RANKINGS\n")
        print(self.vocab)

        self.original_data = self.test_data()
        input = self.input_format()
        for _ in range(tests):
            self.adjust_rankings(input, self.original_data)
            print("\n\n\n\nADJUSTED DATA\n")
            print(self.sort_ratings(self.master_rankings))
            self.get_vocab(self.master_rankings)
            print("\n\nADJUSTED RANKINGS\n")
            print(self.vocab)
            print("\n\n")
            input = self.test_input()
            print("\n\nnew input")
            print(input)



class Core:

    def __init__(self, state=5, vocab=[], constants={"a": 3, "b": 5, "c": 7, "d": 11}, mod=9973, compound=None):

        if isinstance(state, int) == False:
            self.current_state = [int(x) % mod for x in state]
            self.original_state = [int(x) % mod for x in state]
        else:
            self.current_state = state
            self.original_state = state
    
        self.current_vocab = vocab
        self.original_vocab = vocab
        self.constants = constants
        self.mod = mod
        try:
            compound[0] = [compound[0]] if isinstance(compound[0], int) == True else compound[0] if compound else None
        except:
            pass
        self.compound = compound
        print(compound)
        
       
        self.iteration = 1 

        self.prior_index = 0

        self.seed_format = {
            "state": self.current_state,
            "vocab": self.original_vocab,
            "constants": self.constants,
            "mod": self.mod
        }

        if compound:
            self.seed_format["compound"] = compound

        self.token_array = []




    def step(self): 
        self.morph_array()
        self.iteration += 1
        return self.get_token() if not self.compound else self.compound_token()


    def reverse(self):
        if self.iteration > 1:
            value = self.current_state[len(self.current_state)-1]
            self.iteration = self.iteration - 1
            self.current_state = self.current_state[:len(self.current_state)-1] 
            old_value = [self.reverse_morph(value)]
            self.current_state = old_value + self.current_state
        return self.get_token() if not self.compound else self.compound_token()


    def get_token(self):
        return self.current_vocab[self.index_token()] if len(self.current_vocab) > 0 else self.current_state[0]
    

    def generate_tokens(self, amount):
        self.rebuild()
        return [self.step() for _ in range(amount)]
    

    def index_token(self):
        return self.current_state[0] * (self.constants["a"] * (self.constants["c"] + self.constants["b"])) % len(self.current_vocab)


    def morph_array(self):
        value = self.current_state[0]
        self.current_state = self.current_state[1:]
        self.current_state.append(self.new_value(value))


    def reverse_morph(self, value):
        return ((value - self.constants["d"]) * pow((round((self.constants["c"] * self.constants["c"]) / self.constants["a"]) * self.constants["b"] * self.iteration) % self.mod, -1, self.mod)) % self.mod


    def new_value(self, value):
        return self.morph_logic(value) 


    def morph_logic(self, value):
        return (self.constants["d"] + round((self.constants["c"] * self.constants["c"])/self.constants["a"]) * self.constants["b"] * self.iteration * value) % self.mod
    

    def jump(self, index):
        self.rebuild()
        tokens = [self.step() for _ in range(index)]
        return tokens[len(tokens)-1]


    def rebuild(self): 
        self.iteration = 1
        self.current_state = self.original_state


    def compound_token(self):
        count = 0
        tokens = []

        current = self.compound[0][self.current_state[0] * (self.constants["c"] if self.current_state[0] % 3 == 0 else self.constants["b"]) % len(self.compound[0])]
        group = self.current_vocab[(self.current_state[0] * self.constants["b"] if self.current_state[0] * self.current_state[0] * self.constants["d"] % 3 == 0 else self.current_state[0] * self.constants["d"]) % len(self.current_vocab)]
        
        while (len(tokens) < current and current < len(group)) or (len(tokens) < len(group) and current >= len(group)):
            token = group[(self.current_state[0] + count + self.iteration * (len(tokens)*self.constants["c"]) * (self.constants["a"] * (self.constants["c"] + self.constants["b"]))) % len(group)] 
            if token in tokens:
                count += 1
            else:
                tokens.append(token) 
        i = 1
        while i < len(tokens):
            tokens.insert(i, self.compound[2][((len(tokens) + 1) * self.iteration * self.constants["c"]) % len(self.compound[2])])
            i += 2

        return " ".join(tokens)


class Modify:


    def __init__(self, data, **kwargs):
        
        self.edit = None
        self.result = None
        self.data = data


    def get_keys(self):
        items = [list(i.items()) for i in self.data["input"]]

        data = [[str(i[0][1]), i[1][1]] for i in items] if self.data["rules"] != "random" else [[str(i[0][1])] for i in items]

        edit = {} 

        for item in data:
            edit[item[0]] = item[1] if self.data["rules"] != "random" else item[0]

        self.edit = edit


    def modify(self, key, rating):
        self.edit[key] = rating


    def adjust(self, key, adjust):
        self.edit[key] += adjust


    def export(self):
        result = [] 
        
        for item in self.edit:
            result.append({"data": item, "rating": self.edit[item]})

        self.result = result
        return result


import random
from stateshaper.tools.tiny_state.TinyState import TinyState
from .Vocab import Vocab
from .Modify import Modify




class Connector:


    def __init__(self, data, token_count=None, initial_state=None, constants=None, vocab=None, mod=None, **kwargs):
        
        if data["rules"] == "rating":
            data["length"] = len(data["input"]) if data["length"] > len(data["input"]) else data["length"]

        self.debug = True
        self.modify = Modify(data)
        self.tiny_state = TinyState()

        self.default_token_count = 10
        self.default_state = 5
        self.default_constants = {
            "a": 3,
            "b": 5,
            "c": 7,
            "d": 11
        }
        self.default_mod = 9973

        self.engine = None
        self.token_count = self.default_token_count if not token_count else token_count
        self.state = self.initial_state = self.default_state if not initial_state else initial_state
        self.vocab = vocab if vocab else None
        self.compressed_vocab = None
        self.constants = constants if constants else self.default_constants
        self.mod = mod if mod else self.default_mod

        self.compressed_seed = None

        self.check_input(data, constants)

        self.check_random(data)

        self.check_compound(data)

        self.data = data

        self.vocab_rules = {
            "rating": self.get_personalization,
            "random": self.get_data,
            "compound": self.get_data,
            "tokens": self.get_tokens
        }



    def start_connect(self, format=None):
        self.build_seed()
        
        self.engine = self.minimal_seed = {
            "state": self.state,
            "vocab": self.vocab,
            "constants": self.constants,
            "mod": self.mod
        }

        print("\n\n\nStateshaper Seed has been created:\n")
        print(self.engine)

        print("\n\n\nVocab is Compressed:\n")
        print(self.compressed_vocab)

        print("\n\n\nFull Seed:\n")
        self.engine["vocab"] = self.compressed_vocab if format else self.engine["vocab"]

        self.output_seed()

        print("\n\n\nMinimal Output Seed:\n")
        self.compressed_seed["v"] = self.compressed_vocab
        print(self.compressed_seed)
        print("\nSize: " + str(len(str(self.compressed_seed))) + " bytes\n\n\n")
        
        return self.engine
    


    def output_seed(self, state=None, vocab=None, constants=None, mod=None, derived=None):
        seed = {}

        seed["s"] = self.engine["state"] if not state else state
        seed["v"] = self.compressed_vocab if vocab == True else derived if derived else ""
        seed["c"] = self.engine["constants"] if not constants else constants
        seed["m"] = self.engine["mod"] if not mod else mod

        self.compressed_seed = seed  

        return seed


    def check_input(self, data, constants):
        if data and isinstance(data, dict) == False and isinstance(data, list) == False: 
            print("\nData input is invalid. The input requires 'dict' or 'list' format.")

        if constants and (isinstance(constants, dict) == False or len([constants[i] for i in list(constants.keys()) if isinstance(constants[i], int) == False]) > 0):
            print("\nConstants parameter is invalid. It needs to be a dict with keys containing integer values.")

        try:
            isinstance(data["length"], int)
        except:
            data["length"] = 10



    def check_random(self, data):
        try:
            isinstance(data["rules"], str) and data["rules"] == "random"
        except:
            data["rules"] = "random"
            return True
        
        try:
            isinstance(data["modifier"], int)
        except:
            data["modifier"] = 21
            

    def check_compound(self, data):
        try:
            isinstance(data["rules"], str) and data["rules"] == "compound"
        except:
            data["rules"] = "random"
            return True
        
        try:
            isinstance(data["compound_modifier"], int)
        except:
            data["compound_modifier"] = 7

        try:
            isinstance(data["compound_length"], int) or len([i for i in data["compound_length"] if isinstance(i, int) == True]) == len(data["compound_length"])
        except:
            data["compound_length"] = 3

        try:
            isinstance(data["compound_groups"], list)
        except:
            data["compound_groups"] = None

        try:
            isinstance(data["compound_terms"], list)
        except:
            data["compound_terms"] = [" "]


    def build_seed(self):
        if self.vocab:
            self.compressed_vocab = self.vocab
            self.vocab = self.vocab_rules[self.data["rules"]](self.vocab[0], self.vocab[1], self.data) if self.data["rules"] != "tokens" else []
        else:
            self.vocab = self.get_vocab() if self.data["rules"] != "tokens" else []
            self.compressed_vocab = self.compress_vocab() if self.data["rules"] != "tokens" else []
        self.state = self.initial_state
        self.constants = self.constants
        self.mod = self.mod


    def get_constants(self):
        values = ["a", "b", "c", "d"]
        constants = {}
        assigned = 1
        for value in values:
            constant = random.randint(assigned + 1, assigned + 10)
            constants[value] = constant
            assigned = constant

        return constants


    def get_vocab(self):
        if self.vocab:
            return True
        if isinstance(self.data, dict):
            self.vocab_obj = Vocab(self.data)  
            return self.vocab_obj.define_vocab() 
        else:   
            print("no valid data")


    def get_state(self):
        return self.state


    def assign_constants(self, constants=None):
        new_constants = {}
        for key in self.default_constants.keys():
            new_constants[key] = constants[len(new_constants)] if constants else None
        return new_constants if constants else self.default_constants
        

    def get_mod(self):
        if not self.mod:
            self.mod = self.default_mod  


    def change_data(self, data):
        self.data = data


    def change_token(self, token):
        self.token_count = token


    def set_value(self, key, rating):
        self.modify.modify(key, rating)


    def adjust_value(self, key, adjust):
        self.modify.adjust(key, adjust)


    def compress_vocab(self, vocab=None):
        return self.tiny_state.get_seed(self.data, self.vocab if not vocab else vocab)


    def compress_regular(self, vocab=None):
        return self.tiny_state.derived_seed(self.data, self.vocab if not vocab else vocab)
    

    def get_minimal(self):
        return self.compressed_seed
    

    def decode_seed(self, seed):
        return self.tiny_state.decode(seed)


    def decode_minimal(self, seed, minimal):
        return self.tiny_state.decode_subset_seed(seed, minimal)


    def get_personalization(self, seed, minimal, data):
        return self.tiny_state.rebuild_data(seed, minimal, data)


    def get_data(self, seed, minimal, data):
        return self.tiny_state.rebuild_regular(seed, minimal, data)
    

    def get_tokens(self, seed, minimal, data):
        return []


import sys
from stateshaper.connector.Connector import Connector
from stateshaper.core import Core
from stateshaper.tools.derive_vocab.DeriveVocab import DeriveVocab
from stateshaper.tools.tiny_state.TinyState import TinyState
from example_data.format_data.FormatData import FormatData



class Stateshaper:

    def __init__(self, data=None, original_data=None, seed=None, token_count=10, initial_state=5, vocab=None, constants={"a": 3,"b": 5,"c": 7,"d": 11}, mod=9973, **kwargs):
        

        if isinstance(data, dict):
            self.data = data
        else:
            print("Data is not formatted or formatted incorrectly. See accepted data formats in the 'example_data' directory.")
            sys.exit()

        self.original_data = original_data
        
        self.seed = None

        self.first_run = True if not seed else False

        self.token_count = token_count


        if isinstance(initial_state, int):
            initial_state = [initial_state]


        if isinstance(seed, dict):
            try:
                initial_state = seed["s"] 
            except:
                initial_state = initial_state 
            try:
                vocab = seed["v"]
            except:
                vocab = vocab
            try:
                constants = seed["c"]
            except:
                constants = constants
            try: 
                mod = seed["m"] 
            except: 
                mod = mod
        
            self.seed = {
                "state": initial_state,
                "vocab": vocab,
                "constants": constants,
                "mod": mod
            }

        self.connector = Connector(data=self.data, token_count=token_count, initial_state=initial_state,  constants=constants, vocab=vocab, mod=mod)

        self.tiny_state = TinyState()

        self.format_data = FormatData()

        self.derive_vocab = DeriveVocab()

        self.engine = None

        self.default_seed = None

        self.derived_seed = None




    def start_engine(self):
        self.engine = None
        self.seed = self.connector.start_connect()
        self.define_engine()


    def define_engine(self, state=None, vocab=None, constants=None, mod=None):
        self.engine = Core(
            self.seed["state"] if not state else state,
            self.seed["vocab"] if not vocab else vocab,
            self.seed["constants"] if not constants else constants,
            self.seed["mod"] if not mod else mod,
            [self.data["compound_length"], self.data["compound_groups"], self.data["compound_terms"]] if self.data["rules"] == "compound" else None
        )


    def run_engine(self, token_count=None):
        self.tokens = self.engine.generate_tokens(self.token_count if not token_count else token_count)
        print("\n\nTokens successfully generated from vocab.")
        print("============================================")
        print(self.tokens)

        return self.tokens
    

    def reverse(self, token_count=None):
        self.tokens = [self.engine.reverse() for _ in range(self.token_count if not token_count else token_count)]
        print("\n\nToken re-created from reverse.")
        print("============================================")
        print(self.tokens)

    
    def jump(self, index):
        self.tokens = self.engine.jump(index)
        print(f"\n\nToken successfully accessed from index {str(index)}.")
        print("============================================")
        print(self.tokens)


    def one_token(self):
        self.tokens = self.engine.step()
        print("\n\nOne token has been created.")
        print("============================================")
        print(self.tokens)
        return self.tokens 
    

    def reverse_one(self):
        self.tokens = self.engine.reverse() 
        print("\n\nOne token re-created from reverse.")
        print("============================================")
        print(self.tokens)
        return self.tokens


    def rebuild(self):
        self.engine.rebuild()


    def get_seed(self, state=None, vocab=False, constants=None, mod=None):
        return self.connector.output_seed(state=state, vocab=vocab, constants=constants, mod=mod)
    
    
    def get_derived(self):
        return self.derived_seed
    

    def adjust_ratings(self):
        for _ in range(20):
            self.derive_vocab.adjust_rankings(self.original_data, self.data)
        new_vocab = self.derive_vocab.current_vocab()
        self.seed["vocab"]  = new_vocab
        self.derived_seed = self.new_seed(new_vocab)
        self.define_engine()


    def new_seed(self, vocab):
        seed = {}

        seed["s"] = self.seed["state"]  
        seed["v"] = self.connector.compress_regular(vocab)
        seed["c"] = self.seed["constants"]
        seed["m"] = self.seed["mod"]

        self.derived_seed = seed["v"]  

        return seed



import json
from tools.derive_vocab.DeriveVocab import DeriveVocab
from stateshaper import Stateshaper
from tools.tiny_state.TinyState import TinyState


class Demo:


    def __init__(self, user_id=None, **kwargs):
        super().__init__(**kwargs)


        # self.connector = None

        self.run_test("random")
        # DeriveVocab()
        # print(Functions().groups("pizza", "example_data/compound.json"))



    def run_test(self, test="compound"):
        file = self.get_file(test)

        with open(file, "r") as f:
            self.data = json.loads(f.read())
            f.close()

        self.engine = Stateshaper(self.data, token_count=50, initial_state=1)
        self.engine.start_engine()
        self.engine.run_engine()


    def get_file(self, test):
        return f"example_data/{test}.json"



Demo()


from random import randint


class TinyState:

    def __init__(self, list_count=3, **kwargs):
        

        self.subset_alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

        self.list_count = list_count
        self.subset_size = list_count * 4

        self.data = None


    def set_count(self, count):
        self.list_count = count
        self.subset_size = count * 4


    def _encode_letters_from_int(self, n: int) -> str:
        """Encode 0 <= n < 26^3 into three uppercase letters."""
        if not (0 <= n < 26**3):
            raise ValueError("Letter block out of range")
        a = n // (26 * 26)
        b = (n // 26) % 26
        c = n % 26
        return "".join(chr(65 + x) for x in (a, b, c))


    def _decode_letters_to_int(self, letters: str) -> int:
        """Decode three uppercase letters into an integer."""
        if len(letters) != 3 or not all("A" <= ch <= "Z" for ch in letters):
            raise ValueError("Letters must be 3 A–Z characters")
        return ((ord(letters[0]) - 65) * 26 + (ord(letters[1]) - 65)) * 26 + (ord(letters[2]) - 65)


    def _encode_params(self, num_keys: int, num_items: int) -> str:
        """
        Pack (num_keys, num_items) into 'ABC12345'.

        We assume:
          1 <= num_keys <= 99
          1 <= num_items <= 99

        Encoding:
          param_index = num_keys * 100 + num_items   (max 9900)
          block_int   = param_index // 100000        (0 for these sizes)
          num_int     = param_index % 100000
          seed        = ABC + 5-digit decimal
        """
        if not (1 <= num_keys <= 99 and 1 <= num_items <= 99):
            raise ValueError("num_keys / num_items out of supported range (1–99)")

        param_index = num_keys * 100 + num_items  # max 9900

        block_int = param_index // 100000         # will be 0 here
        num_int = param_index % 100000

        letters = self._encode_letters_from_int(block_int)  # 'AAA' for now
        digits = f"{num_int:05d}"                           # 00000..99999
        return letters + digits                             # e.g. 'AAA01005'
    

    def _decode_params(self, seed: str) -> tuple[int, int]:
        """
        Inverse of _encode_params: 'ABC12345' -> (num_keys, num_items).

        NOTE:
          This only encodes the LAYOUT (grid shape), not any sparse subset.
        """
        # if len(seed) != 8:
        #     raise ValueError("Seed must be exactly 8 characters: 'ABC12345'")

        letters = seed[:3]
        digits = seed[3:]

        if not digits.isdigit():
            raise ValueError("Last 5 characters must be digits")

        block_int = self._decode_letters_to_int(letters)
        num_int = int(digits)

        param_index = block_int * 100000 + num_int
        num_keys = param_index // 100
        num_items = param_index % 100

        if num_keys <= 0 or num_items <= 0:
            raise ValueError("Decoded invalid num_keys / num_items")

        return num_keys, num_items


    def _infer_grid_dimensions(self, s: str) -> tuple[int, int]:
        """
        From a long 'DDIIDDIIDDII...' string, infer:
          - number of distinct dict indices (num_keys)
          - number of distinct item indices (num_items)

        STRICT MODE — Assumes:
          - length is multiple of 4
          - each 4-char block is 'DDII' with 2-digit decimal ints
          - the sequence enumerates ALL (dict_idx, item_idx) in row-major:
                dict_idx 0..num_keys-1
                item_idx 0..num_items-1
            in order:
                (0,0),(0,1)...(0,num_items-1),
                (1,0)...(num_keys-1,num_items-1)
        """
        if len(s) % 4 != 0:
            raise ValueError("Input length must be a multiple of 4 (DDII blocks)")

        blocks = [s[i:i+4] for i in range(0, len(s), 4)]
        pairs = [(int(b[:2]), int(b[2:])) for b in blocks]

        dict_indices = sorted({d for d, _ in pairs})
        item_indices = sorted({i for _, i in pairs})

        if not dict_indices or dict_indices[0] != 0 or dict_indices != list(range(dict_indices[-1] + 1)):
            raise ValueError("Dict indices are not contiguous starting at 0")
        if not item_indices or item_indices[0] != 0 or item_indices != list(range(item_indices[-1] + 1)):
            raise ValueError("Item indices are not contiguous starting at 0")

        num_keys = len(dict_indices)
        num_items = len(item_indices)

        expected = [(d, i) for d in range(num_keys) for i in range(num_items)]
        if expected != pairs:
            raise ValueError("Sequence is not in row-major (dict, item) order")

        return num_keys, num_items


    def _subset_from_original(self, original: str, new: str) -> str:
        """
        Return the subset of DDII blocks from `original` that are present in `new`,
        keeping the order they appear in `original`.

        Both strings must be concatenations of 4-char 'DDII' blocks (len % 4 == 0).
        """
        if len(original) % 4 != 0 or len(new) % 4 != 0:
            raise ValueError("Strings must be multiples of 4 (DDII blocks)")

        orig_blocks = [original[i:i+4] for i in range(0, len(original), 4)]
        new_blocks = set(new[i:i+4] for i in range(0, len(new), 4))

        subset_blocks = [blk for blk in orig_blocks if blk in new_blocks]
        return "".join(subset_blocks)


    def compress(self, big_string: str) -> str:
        """
        Compress a long STRICT 'DDIIDDIIDDI...' string into 'ABC12345'.

        This expects a FULL contiguous grid in row-major order:
          - dict_idx  0..num_keys-1
          - item_idx  0..num_items-1
          - all positions present
        """
        num_keys, num_items = self._infer_grid_dimensions(big_string)
        return self._encode_params(num_keys, num_items)


    def decode(self, seed: str) -> str:
        """
        Decode 'ABC12345' back into the full STRICT 'DDIIDDIIDDI...' grid string.

        Uses the encoded (num_keys, num_items) and regenerates
        the canonical row-major order:
            for dict_idx in 0..num_keys-1:
                for item_idx in 0..num_items-1:
                    append f"{dict_idx:02d}{item_idx:02d}"
        """
        num_keys, num_items = self._decode_params(seed)

        parts = []
        for d in range(num_keys):
            for i in range(num_items):
                parts.append(f"{d:02d}{i:02d}")
        return "".join(parts)


    def _build_block_index(self, full_grid: str) -> list[str]:
        """
        Split a full-grid DDII string into blocks.

        Returns list of 4-char blocks in order. Used as the index for bitmask.
        """
        if len(full_grid) % 4 != 0:
            raise ValueError("full_grid must be a multiple of 4 characters")

        blocks = [full_grid[i:i+4] for i in range(0, len(full_grid), 4)]
        return blocks

    def encode_subset_seed(self, full_grid: str, subset: str) -> str:
        """
        Encode a sparse DDII subset as positions within the full grid,
        using a single character per position from subset_alphabet.

        - full_grid: canonical DDII full grid from decode(layout_seed)
        - subset:    sparse DDII string (blocks taken from full_grid)

        Example:
            full_grid blocks indices: 0..39
            positions [0,1,2,7,37,39] might encode as: '012Hb d'
            (depending on alphabet; with the alphabet below, it's '012Hb d'
             but without spaces: '012Hbd')
        """
        if len(full_grid) % 4 != 0 or len(subset) % 4 != 0:
            raise ValueError("Strings must be multiples of 4 (DDII blocks)")

        full_blocks = [full_grid[i:i+4] for i in range(0, len(full_grid), 4)]
        subset_blocks = [subset[i:i+4] for i in range(0, len(subset), 4)]

        n = len(full_blocks)
        if n > len(self.subset_alphabet):
            raise ValueError(
                f"Subset alphabet only supports {len(self.subset_alphabet)} positions, "
                f"but full_grid has {n}"
            )

        index_map = {blk: idx for idx, blk in enumerate(full_blocks)}

        chars = []
        for blk in subset_blocks:
            if blk not in index_map:
                raise ValueError(f"Subset block {blk!r} not found in full_grid")
            pos = index_map[blk]
            chars.append(self.subset_alphabet[pos])

        return "".join(chars)
    

    def decode_subset_seed(self, full_grid: str, compressed_subset: str) -> str:
        """
        Decode a subset seed created by encode_subset_seed back into a sparse
        DDII string, using the same full_grid.

        - full_grid: canonical DDII full grid from decode(layout_seed)
        - subset_seed: string of characters from subset_alphabet
        """
        if len(full_grid) % 4 != 0:
            raise ValueError("full_grid must be a multiple of 4 characters")

        full_blocks = [full_grid[i:i+4] for i in range(0, len(full_grid), 4)]
        n = len(full_blocks)
        if n > len(self.subset_alphabet):
            raise ValueError(
                f"Subset alphabet only supports {len(self.subset_alphabet)} positions, "
                f"but full_grid has {n}"
            )

        sparse_blocks = []
        for ch in compressed_subset:
            try:
                pos = self.subset_alphabet.index(ch)
            except ValueError:
                raise ValueError(f"Invalid subset seed character {ch!r}")

            if pos >= n:
                raise ValueError(
                    f"Subset index {pos} out of range for full_grid size {n}"
                )

            sparse_blocks.append(full_blocks[pos])

        return "".join(sparse_blocks)
    

    def decode_subset(self, layout_seed: str, compressed_subset: str) -> str:
        """
        High-level API:
          - layout_seed: ABC12345 (grid shape, used to build full_grid)
          - compressed_subset: positions-only string encoded with letters/digits

        Returns:
          Reconstructed sparse DDII string.
        """
        full_grid = self.decode(layout_seed)  
        return self.decode_subset_seed(full_grid, compressed_subset)
    
    
    def set_preferences(self, data, length=5):
        self.top_preferences = data[:self.data["length"]]


    def sort_ratings(self, data, length=5):
        sort = sorted(data["input"], key=lambda x: list(x.values())[0]["rating"], reverse=True)
        return [list(i.keys())[0] for i in sort]
    

    def get_seed(self, data, vocab=None, derived=None):


        self.set_count(data["length"])

        self.data = data
        export = []
        partial = []
        side = []
        seed = ""
        subseed = ""

        if data["rules"] == "rating":
            input = self.sort_ratings(data)

            self.set_preferences(input)

            for item in data["input"]:
                key = list(item.keys())[0]
                for term in item[list(item.keys())[0]]["data"]:
                    idx1 = data["input"].index(item)
                    idx2 = item[list(item.keys())[0]]["data"].index(term)
                    if len(export) < self.data["length"]:
                        if len([x for x in term["attributes"] if x in self.top_preferences and key == self.top_preferences[0]]) > 0:
                            export.append(f"{idx1:02d}{idx2:02d}")
                        elif len([x for x in term["attributes"] if x in self.top_preferences]) > 0 and len([y for y in self.top_preferences if key == y]) > 0:
                            partial.append(f"{idx1:02d}{idx2:02d}")
                        else:
                            side.append(f"{idx1:02d}{idx2:02d}")

                    seed = seed + f"{idx1:02d}{idx2:02d}"
                    
            
            subseed = export + partial + side

            subseed = subseed[:data["length"]]

            subseed = "".join(subseed)

            self.seed = subseed

            self.original_seed = seed 

            self.compressed_seed = self.compress(seed)

            self.compressed_subset = self.encode_subset_seed(seed, subseed)

            self.decoded_subset = self.decode_subset_seed(seed, self.compressed_subset)

            print("\n\n\nFull list based on ratings profile:\n")
            print(vocab)

            print("\n\n\nCompressed Tiny State format for entire list:\n")
            print(self.compressed_seed)

            print("\n\n\nCompressed seed for chosen data set:\n")
            print(self.compressed_subset)

            print("\n\n\n List rebuilt from extracted seed:\n")
            print(self.rebuild_data(self.compressed_seed, self.compressed_subset, self.data))
            print("\n\n")
            
            print("\nCompare to final list:\n")
            self.exported_data = vocab
            print(vocab)
            print("\n\n")

            return [self.compressed_seed, self.compressed_subset]
        
        else:
            return self.regular_vocab(data, vocab)


    def regular_vocab(self, data, vocab):
        seed = ""
        subseed = ""
        for item in data["input"]:
            idx1 = data["input"].index(item)
            idx2 = 00
            seed = seed + f"{idx1:02d}{idx2:02d}"

        print("\n\n\nOriginal List:\n")
        print(vocab)

        self.original_seed = seed
        print("\n\n\nOriginal Seed:\n")
        print(self.original_seed)

        self.compressed_seed = self.compress(self.original_seed)

        print("\n\n\nCompressed Tiny State format for entire list:\n")
        print(self.compressed_seed)

        for item in data["input"]:
            if item["data"] in vocab:
                idx1 = data["input"].index(item)
                idx2 = 00 
                subseed = subseed + f"{idx1:02d}{idx2:02d}"

        print("\n\n\nOriginal Subseed:\n")
        print(subseed)

        self.compressed_subset = self.encode_subset_seed(seed, subseed)

        print("\n\n\nCompressed seed for chosen data set:\n")
        print(self.compressed_subset)

        self.decoded_subset = self.decode_subset_seed(seed, self.compressed_subset)

        print("\n\n\nList rebuilt from extracted seed:\n")
        
        print(self.rebuild_regular(self.compressed_seed, self.compressed_subset,  self.data))
        print("\n\n")
        
        return [self.compressed_seed, self.compressed_subset]


    def derived_seed(self, data, vocab=None):
        self.set_count(data["length"])
        self.data = data
        export = []
        side = []
        seed = ""
        subseed = ""

        if data["rules"] == "rating":

            for item in data["input"]:

                idx1 = data["input"].index(item)
                idx2 = 00
                try:
                    term = list(item.values())[0]["data"][0]["item"]
                except:
                    term = list(item.values())[0]["data"]["item"]
                if len(export) < self.data["length"]:
                    if term in vocab:
                        print(term)
                        export.append(f"{idx1:02d}{idx2:02d}")
                    else:
                        side.append(f"{idx1:02d}{idx2:02d}")

                seed = seed + f"{idx1:02d}{idx2:02d}"
                    
            
            subseed = export 

            subseed = subseed[:data["length"]]

            subseed = "".join(subseed)

            self.seed = subseed

            self.original_seed = seed 

            self.compressed_seed = self.compress(seed)

            self.compressed_subset = self.encode_subset_seed(seed, subseed)

            self.decoded_subset = self.decode_subset_seed(seed, self.compressed_subset)

            print("\n\n\nFull list based on ratings profile:\n")
            print(vocab)

            print("\n\n\nDerived Tiny State format for entire list:\n")
            print(self.compressed_seed)

            print("\n\n\nDerived seed for chosen data set:\n")
            print(self.compressed_subset)

            print("\n\n\nDerived list rebuilt from extracted seed:\n")
            print(self.rebuild_derived(self.compressed_seed, self.compressed_subset, self.data))
            print("\n\n")
            
            print("\nCompare to final list:\n")
            self.exported_data = vocab
            print(vocab)
            print("\n\n")

            return [self.compressed_seed, self.compressed_subset]
        
        else:
            return self.regular_vocab(data, vocab)
        

    def rebuild_data(self, compressed_seed, compressed_subset, data=None):
        origin_seed = self.decode(compressed_seed)
        decoded = self.decoded = self.decode_subset_seed(origin_seed, compressed_subset) 
        export = []

        while len(export)<data["length"]:
            parent = decoded[:2]
            child = decoded[2:4]
            key = list(data["input"][int(parent)].keys())[0]
            export.append(data["input"][int(parent)][key]["data"][int(child)]["item"])
            decoded = decoded[4:]

        return export
    
    
    def rebuild_derived(self, compressed_seed, compressed_subset, data=None):
        origin_seed = self.decode(compressed_seed)
        decoded = self.decoded = self.decode_subset_seed(origin_seed, compressed_subset) 
        export = []

        while len(export)<data["length"]:
            parent = decoded[:2]
            child = 00
            key = list(data["input"][int(parent)].keys())[0]
            export.append(data["input"][int(parent)][key]["data"][int(child)]["item"])
            decoded = decoded[4:]

        return export
    

    def rebuild_regular(self, compressed_seed, compressed_subset, data=None):
        origin = self.decode(compressed_seed)
        decoded = self.decoded = self.decode_subset_seed(origin, compressed_subset) 
        export = []
        while len(decoded) > 0:
            index = decoded[:2]
            export.append(data["input"][int(index)]["data"])
            decoded = decoded[4:]

        return export


from random import randint
from stateshaper.tools.derive_vocab.DeriveVocab import DeriveVocab


class Vocab:


    def __init__(self, data, **kwargs):
        

        self.rule_types = None
        self.debug = True
        self.data = data 
        self.derive_vocab = DeriveVocab()

        try:
            self.random_mods = data["random_mods"]
        except:
            self.random_mods = [10, 7]

        try:
            self.random_constants = data["random_constants"]
        except:
            self.random_constants = [2, 3, 5, 8]

            

        self.rule_types = ["random", "rating", "compound", "token"]

        self.rules_explained = {
            "random": "The vocab consists of any terms within the input. If the length value is define_vocab, the vocab list will include that many random values from the initial input data.",
            "rating": "Create the vocab based on 1-100 ratings. If the length value is define_vocab, the vocab list will include that many items from the input data, based on the highest ratings. If no length is specified, the vocab will be all input.",
            "compound": "Items from the vocab list will be randomly combined based on the 'group' value and called during each iteration of Stateshaper engine. If a length is specified, the vocab list will be limited to that many items. If an compound_groups list is set, only those groups will be in the list, otherwise any group in the input data can be included.",

            # IN PROGRESS
            "token": "The vocab list will consist of objects or functions that are called during Stateshaper iterations. This will be based on a number ranking. If a length value is specified, the vocab list will be limited to that number. "
        }

        self.compound_rules = ["random"] 

        self.mapping_types = {
            "random": self.random_mapping,
            "rating": self.rating_mapping,
            "compound": self.compound_mapping,
            "token": self.token_mapping 
        }

        self.current_rule = self.data["rules"] if self.valid_rule(self.data["rules"]) else self.rule_types[0]
        self.data_map = {}
        self.vocab = []


        

    def set_data(self, data):
        self.data = data
        self.valid_data()
        self.define_vocab(self.data["rules"])


    def valid_data(self):
        if self.data["rules"] and self.valid_rule(self.data["rules"]) == False:
            print("The rule chosen is not valid. Valid types are 'rating', 'compound', 'random', or 'token'. The rule will be defaulted to 'random'.")
            self.data["rules"] = "random"

        if (self.data["rules"] == "compound" and self.data["compound_length"] and (isinstance(self.data["compound_length"], int) == False) or len([i for i in self.data["compound_length"] if isinstance(i, int) == True]) < len(self.data["compound_length"])) or not self.data["compound_length"]:
            print("\n\nThe 'compound' rule has been chosen, but no compound length has been assigned. Using default compound length of 2.")
            self.data["compound_length"] = 3

        if (self.data["rules"] == "compound" and self.data["compound_rules"] and (isinstance(self.data["compound_rules"], int) == False or self.valid_compound_rules() == False)) or not self.data["compound_rules"]:
            print("\n\nThe 'compound' rule has been chosen, but no compound rules have been assigned, or the assigned rule is invalid. Using default compound rule'random'.")
            self.data["compound_rules"] = "random"

        if self.data["rules"] == "compound" and (self.data["compound_groups"] and  self.valid_compound_groups() == False) or not self.data["compound_groups"]:
            print("\n\nThe 'compound' rule has been chosen, but no compound rules have been assigned, or the assigned rule is invalid. Using default compound rule'random'.")
            self.data["compound_rules"] = "random"

        if self.data["length"] and isinstance(self.data["length"], int) == False:
            print("The length value is not an integer. Length will be set to input list size.")
            self.data["length"] = len(self.data["input"])

        print("\n\n")

        if isinstance(self.data, dict) == False:
            print("Data passed is in incorrect format. Please make sure it is a dict/json with keys 'input' (list) and 'rules' (string).")
            
        elif isinstance(self.data["input"], list) == False:
            print("The input list isn't formatted correctly. Please ensure it is formatted as a list/array.")
            
        elif self.data["rules"] == "random" and self.valid_random() == False:
            print("The 'random' rule has been chosen, but the input list is invalid. Please make sure each value in the input data set is in the following format:\n\n{'data': values here}.")

        elif self.data["rules"] == "rating" and self.valid_ratings() == False:
            print("The 'rating' rule has been chosen, but the input list is invalid. Please make sure each value in the input data set is in the following format:\n\n{'data': values here, 'rating': 0/100 rating here}.")

        elif self.data["rules"] == "compound" and self.valid_compound() == False:
            print("The 'compound' rule has been chosen, but the input list is invalid. Please make sure each value in the input data set is in the following format:\n\n{'data': values here, 'group': your specified group type, int or str}.")

        elif self.data["rules"] == "token" and self.valid_tokens() == False:
            print("The 'token' rule has been chosen, but not all rules are define_vocab. Please make sure each value in the input data set is in the following format:\n\n{'data': use an object, function call etc to call base on the value in the Stateshaper array, 'rank': int in event calling order preference (1, 2, 3).")
      
        else:
            print("The data has been accepted. Processing input to enter into Stateshaper...")
            self.define_vocab() if not self.data["rules"]  and not self.debug else self.define_vocab(self.data["rules"]) if not self.debug else self.test()

        print("\n\n")


    def valid_rule(self, rule):
        if rule not in self.rule_types:
            print("\nRule in data set is not in list of define_vocab rules.")
        return rule in self.rule_types
    

    def valid_ratings(self):
        i = 0
        for item in self.data["input"]:
            i += 1
            if "data" not in item.keys() or "rating" not in item.keys() and len(item.keys()) == 2:
                print(f"Bad keys in object {i}.\n")
                return False
            if len(item.keys()) > 2:
                print(f"Too many keys in object {i}.\n")
                return False
            if isinstance(item["rating"], int) == True:
                if item["rating"] > 100 or item["rating"] < 0:
                    print(f"Rating is out of range in object {i}.\n")
                    return False
            if isinstance(item["rating"], int) == False:
                    print(f"Rating is not an integer in object {i}.\n")
                    return False
        print("Input 'rating' data is valid.\n\n")
        print("Length: " + str(self.data["length"]) + "\n\n")
        return True


    def valid_compound(self):
        i = 0
        for item in self.data["input"]:
            i += 1
            if "data" not in item.keys() or "group" not in item.keys() and len(item.keys()) == 2:
                print(f"Bad keys in object {i}.\n")
                return False
            if len(item.keys()) > 2:
                print(f"Too many keys in object {i}.\n")
                return False
            if (isinstance(item["group"], int) == False and isinstance(item["group"], str) == False):
                (f"Group is invalid in object {i}. Please make sure the group id is an int or str variable type.")
                return False
        print("Input 'compound' data is valid.\n\n")
        print("Length: " + str(self.data["length"]) + "\n\n")
        return True
    

    def valid_compound_rules(self):
        return self.data["compound_rules"] in self.compound_rules
    

    def valid_compound_groups(self):
        i = 0
        for item in self.data["input"]:
            i += 1
            if "group" not in item.keys() or not item["group"]:
                print(f"Compound rule is invalid in object {i}.\n")
                return False
        print("Compound groups are valid.\n\n")
        print("Length: " + str(self.data["length"]) + "\n\n")
        return True         
    
 
    def valid_tokens(self):
        i = 0
        for item in self.data["input"]:
            i += 1
            if "data" not in item.keys() or "rank" not in item.keys() and len(item.keys()) == 2:
                print(f"Bad keys in object {i}.\n")
                return False
            if len(item.keys()) > 2:
                print(f"Too many keys in object {i}.\n")
                return False
            if not item["rank"]:
                print(f"Rank value is not valid in object {i}.\n")
                return False
            if item["rank"] and isinstance(item["rank"], int) == False:
                print(f"Rank value is not an integer in object {i}.\n")
                return False
        print("Rule 'token' data is valid.\n\n")
        print("Length: " + str(self.data["length"]) + "\n\n")
        return True      


    def length_exists(self):
        if isinstance(self.data["length"], int):
            self.data["length"] = len(self.data["input"]) if self.data["length"] > len(self.data["input"]) else None

    
    def define_vocab(self):
        self.mapping_method()

        self.print_map()
        return self.vocab

        
    def mapping_method(self):
        self.mapping_types[self.current_rule]()


    def random_mapping(self):
        vocab = [i["data"] for i in self.data["input"] if self.data["input"].index(i) % (self.random_mods[0] if self.data["input"].index(i) % 2 == 0 else self.random_mods[1]) in self.random_constants]
        self.vocab = vocab[:self.data["length"]]


    def set_preferences(self, data, length=5):
        self.top_preferences = self.derive_vocab.initial_rankings(self.data)


    def rating_mapping(self):
        print("\n\nStarting ratings based mapping.\n")
        export = []
        partial = []
        side = []
        full_list = []


        input = self.sort_ratings()

        self.set_preferences(input)

        for interest in self.data["input"]:        
            key = list(interest.keys())[0]
            for item in interest[list(interest.keys())[0]]["data"]:
                if len(export) < self.data["length"]:
                    if len([x for x in item["attributes"] if x in self.top_preferences and key == self.top_preferences[0]]) > 0:
                        export.append(item["item"])
                    elif len([x for x in item["attributes"] if x in self.top_preferences]) > 0 and len([y for y in self.top_preferences if key == y]) > 0:
                        partial.append(item["item"])
                    else:
                        side.append(item["item"])

        full_list = export + partial + side

        full_list = full_list[:self.data["length"]]

        print(f"\nStateshaper vocab parameter successfully set with 'rating' rule.\n")
        self.vocab = full_list


    def sort_ratings(self, length=5):
        sort = sorted(self.data["input"], key=lambda x: list(x.values())[0]["rating"], reverse=True)
        return [list(i.keys())[0] for i in sort]
    

    def compound_mapping(self):
        if self.data["compound_groups"]:
            included = [[item["data"] for item in self.data["input"] if len([x for x in item["groups"] if x in i]) == len(i)] for i in self.data["compound_groups"]]
             
        self.vocab = included

        print(f"\nStateshaper vocab parameter successfully set with 'compound' rule.")


    def token_mapping(self):
        print("\n\nStarting token based mapping. Events will be called based on number ranking.\n")

        included = []
        input = self.sort_token()

        i = 0
        while len(included) < self.data["length"]:
            included.append(input[i]["data"])
            i+=1

        self.vocab = included

        print(f"\nStateshaper vocab parameter successfully set with 'token' rule.")


    def sort_token(self):
        return sorted(self.data["input"], key=lambda x: x["rank"], reverse=False)        


    def print_map(self):
        print("\n\nThe vocab list is set. Items:\n\n")

        for item in self.vocab:
            print(str(item))

        print(f"\n\nThe vocab will be called based on the '{str(self.current_rule)}' rule:\n")
        print(self.rules_explained[self.current_rule])


from setuptools import setup, find_packages

setup(
    name="stateshaper",
    version="0.1.0",
    description="Create infinite deterministic data from a small seed. Compressed. Reproducible. Reversible. Jump to any state from anywhere.",
    author="Jason G. Dunn",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires=">=3.8",
)
